@@ -16,9 +16,10 @@
 // Copyright (C) 2006, 2008 Pino Toscano <pino@kde.org>
 // Copyright (C) 2007, 2010, 2011 Carlos Garcia Campos <carlosgc@gnome.org>
 // Copyright (C) 2008 Hugo Mercier <hmercier31@gmail.com>
-// Copyright (C) 2008-2010, 2012 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2008-2010, 2012-2014, 2016 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2009 Kovid Goyal <kovid@kovidgoyal.net>
 // Copyright (C) 2009 Ilya Gorenbein <igorenbein@finjan.com>
+// Copyright (C) 2012 Tobias Koening <tobias.koenig@kdab.com>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -219,8 +220,8 @@ LinkDest::LinkDest(Array *a) {
       if (obj2.isNull()) {
 	changeZoom = gFalse;
       } else if (obj2.isNum()) {
-	changeZoom = gTrue;
 	zoom = obj2.getNum();
+	changeZoom = (zoom == 0) ? gFalse : gTrue;
       } else {
 	error(errSyntaxWarning, -1, "Bad annotation destination position");
 	goto err1;
@@ -230,30 +231,26 @@ LinkDest::LinkDest(Array *a) {
 
   // Fit link
   } else if (obj1.isName("Fit")) {
-    if (a->getLength() < 2) {
-      error(errSyntaxWarning, -1, "Annotation destination array is too short");
-      goto err2;
-    }
     kind = destFit;
 
   // FitH link
   } else if (obj1.isName("FitH")) {
-    if (a->getLength() < 3) {
-      error(errSyntaxWarning, -1, "Annotation destination array is too short");
-      goto err2;
-    }
     kind = destFitH;
-    a->get(2, &obj2);
-    if (obj2.isNull()) {
+    if (a->getLength() < 3) {
       changeTop = gFalse;
-    } else if (obj2.isNum()) {
-      changeTop = gTrue;
-      top = obj2.getNum();
     } else {
-      error(errSyntaxWarning, -1, "Bad annotation destination position");
-      kind = destFit;
+      a->get(2, &obj2);
+      if (obj2.isNull()) {
+	changeTop = gFalse;
+      } else if (obj2.isNum()) {
+	changeTop = gTrue;
+	top = obj2.getNum();
+      } else {
+	error(errSyntaxWarning, -1, "Bad annotation destination position");
+	kind = destFit;
+      }
+      obj2.free();
     }
-    obj2.free();
 
   // FitV link
   } else if (obj1.isName("FitV")) {
@@ -312,10 +309,6 @@ LinkDest::LinkDest(Array *a) {
 
   // FitB link
   } else if (obj1.isName("FitB")) {
-    if (a->getLength() < 2) {
-      error(errSyntaxWarning, -1, "Annotation destination array is too short");
-      goto err2;
-    }
     kind = destFitB;
 
   // FitBH link
@@ -558,9 +551,11 @@ LinkURI::LinkURI(Object *uriObj, GooString *baseURI) {
       // relative URI
       if (baseURI) {
 	uri = baseURI->copy();
-	c = uri->getChar(uri->getLength() - 1);
-	if (c != '/' && c != '?') {
-	  uri->append('/');
+	if (uri->getLength() > 0) {
+	  c = uri->getChar(uri->getLength() - 1);
+	  if (c != '/' && c != '?') {
+	    uri->append('/');
+	  }
 	}
 	if (uri2->getChar(0) == '/') {
 	  uri->append(uri2->getCString() + 1, uri2->getLength() - 1);
@@ -700,9 +695,10 @@ LinkSound::~LinkSound() {
 //------------------------------------------------------------------------
 
 LinkRendition::LinkRendition(Object *obj) {
-  operation = -1;
+  operation = NoRendition;
   media = NULL;
   js = NULL;
+  int operationCode = -1;
 
   if (obj->isDict()) {
     Object tmp;
@@ -721,25 +717,43 @@ LinkRendition::LinkRendition(Object *obj) {
     tmp.free();
 
     if (obj->dictLookup("OP", &tmp)->isInt()) {
-      operation = tmp.getInt();
-      if (!js && (operation < 0 || operation > 4)) {
-        error(errSyntaxWarning, -1, "Invalid Rendition Action: unrecognized operation valued: {0:d}", operation);
+      operationCode = tmp.getInt();
+      if (!js && (operationCode < 0 || operationCode > 4)) {
+        error(errSyntaxWarning, -1, "Invalid Rendition Action: unrecognized operation valued: {0:d}", operationCode);
       } else {
         Object obj1;
 
         // retrieve rendition object
         if (obj->dictLookup("R", &renditionObj)->isDict()) {
           media = new MediaRendition(&renditionObj);
-	} else if (operation == 0 || operation == 4) {
-          error(errSyntaxWarning, -1, "Invalid Rendition Action: no R field with op = {0:d}", operation);
+	} else if (operationCode == 0 || operationCode == 4) {
+          error(errSyntaxWarning, -1, "Invalid Rendition Action: no R field with op = {0:d}", operationCode);
 	  renditionObj.free();
 	}
 
 	if (!obj->dictLookupNF("AN", &screenRef)->isRef() && operation >= 0 && operation <= 4) {
-	  error(errSyntaxWarning, -1, "Invalid Rendition Action: no AN field with op = {0:d}", operation);
+	  error(errSyntaxWarning, -1, "Invalid Rendition Action: no AN field with op = {0:d}", operationCode);
 	  screenRef.free();
 	}
       }
+
+      switch (operationCode) {
+        case 0:
+          operation = PlayRendition;
+          break;
+        case 1:
+          operation = StopRendition;
+          break;
+        case 2:
+          operation = PauseRendition;
+          break;
+        case 3:
+          operation = ResumeRendition;
+          break;
+        case 4:
+          operation = PlayRendition;
+          break;
+      }
     } else if (!js) {
       error(errSyntaxWarning, -1, "Invalid Rendition action: no OP or JS field defined");
     }
